import argparse
import json
import sys

import yaml
import zmq


# Defining available options
agipd_options = {
        "force-hg-if-below": float,
        "rel_gain": bool,
        "xray_gain": bool,
        "blc-noise": bool,
        "blc-stripes": bool,
        "blc-set-min": bool,
        "dont-zero-nans": bool,
        "dont-zero-orange": bool,
        "max-pulses": list,
        "calfile": str
    }

available_options = {
        "SPB_DET_AGIPD1M-1": agipd_options,
        "MID_DET_AGIPD1M-1": agipd_options}

parser = argparse.ArgumentParser(
    description='Request update of configuration')
required_args = parser.add_argument_group('required arguments')
required_args.add_argument('--karabo-id', type=str,
                           choices=['SPB_DET_AGIPD1M-1',
                                    'MID_DET_AGIPD1M-1'])
required_args.add_argument('--task', type=str, choices=['correct', 'dark'])
required_args.add_argument('--proposal', type=str,
                    help='The proposal number, without leading p, but with leading zeros')  # noqa
required_args.add_argument('--instrument', type=str,
                    choices=["SPB", "MID", "FXE", "SCS", "SQS", "HED",
                             "DETLAB, CALLAB"], help='The instrument')  # noqa
required_args.add_argument('--cycle', type=str, help='The facility cycle')
parser.add_argument('--apply', action='store_true')
# remove help calls as they will cause the argument parser to exit
add_help = False
if "-h" in sys.argv:
    sys.argv.remove("-h")
    add_help = True
if "--help" in sys.argv:
    sys.argv.remove("--help")
    add_help = True

for det, val in available_options.items():
    bool_keys = []
    for k, v in val.items():
        if v == bool:
            bool_keys.append(k)
    for b in bool_keys:
        available_options[det]['no-{}'.format(b)] = bool

known, remaining = parser.parse_known_args()
args = vars(known)
karabo_id = args["karabo_id"]

if karabo_id is not None:
    for option, typ in available_options[karabo_id].items():
        if typ == list:
            nargs = '+'
        else:
            nargs = None
        parser.add_argument(f"--{option}", type=typ, nargs=nargs)

if add_help:
    sys.argv.append("--help")

args = vars(parser.parse_args())

task = args['task']
instrument = args['instrument']
proposal = args['proposal']
cycle = args['cycle']

if task is None or instrument is None or proposal is None or cycle is None:
    print("Need to define all fields")
    exit()

new_conf = {task: {instrument: {karabo_id: {}}}}
for key, value in args.items():
    key = key.replace("_", "-")
    if key in available_options[karabo_id] and value is not None:

        if isinstance(value, list):
            for v in value:
                value[value.index(v)] = ''.join(v)

        if 'no-' in key and isinstance(value, bool):
            if key not in bool_keys:
                new_conf[task][instrument][karabo_id][key.replace('no-','')] = False  #noqa
            # avoid saving the "no-"key into the updated config
            continue
        # Assure adding an empty string for new empty
        # str. updates (e.g. calfile)
        if isinstance(key, str) and (value == '' or value == ' '):
            value = '""'

        new_conf[task][instrument][karabo_id][key] = value

pyaml = yaml.dump(new_conf, default_flow_style=False)

if not args["apply"]:
    print("\n")
    print("-" * 80)
    print("THIS IS A DRY RUN ONLY, NO CHANGES ARE MADE")
    print("\n")
    print("-" * 80)

print(f"Sending the following update: \n {pyaml}")
print("-" * 80)
con = zmq.Context()
socket = con.socket(zmq.REQ)
con = socket.connect("tcp://max-exfl016:5555")
msg = "','".join(["update_conf", "SASEX", args["karabo_id"],
                  args["instrument"], args["cycle"], args["proposal"],
                  json.dumps(new_conf), str(args["apply"])])
socket.send("['{}']".format(msg).encode())
resp = socket.recv_multipart()[0]
print("Configuration now in place is:")
print(resp.decode())
